import graph.Edge;
import graph.Graph;
import graph.GraphInfo;
import graph.Node;
import graph.NodePair;
import graph.TwoNodes;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.TreeSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.SortedSet;
import java.io.File;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

import options.CmdLineParser;

public class CompareGraph
{
    Graph origGraph;
    Graph summaryGraph;
    double eps;
    int summaryCost;

    public static void main(String[] args)
    {
        CompareGraph g = new CompareGraph(args);
        if(g.compare()){
            System.out.println("GRAPHS DIFFER");
        } else {
            System.out.println("GRAPHS ARE SAME");
        }
    }

    public CompareGraph(String[] args){

        long start = System.currentTimeMillis();

		// get options.
		CmdLineParser parser = new CmdLineParser();
		CmdLineParser.Option selfloops = parser.addBooleanOption('s',"self");

		try { parser.parse(args); }
		catch ( CmdLineParser.OptionException e ) {
            System.err.println(e.getMessage());
            System.err.println("Usage: Greedy [-s,--self] graph_file");
            System.exit(2);
        }

		Boolean selfloopsValue = (Boolean) parser.getOptionValue(selfloops);

        //-------GETS ARGS------------
        String filename1 = args[0]; // input data.
        String filename2 = args[1]; // summarized data.
        eps = 0;
        if(args.length > 2)
            eps = Double.valueOf(args[2]);

        System.out.println("Filename: " + filename1);
        System.out.println("Filename: " + filename2);
        System.out.println("Epsilon: " + eps);

        //----------READ INPUT FILE, MAKE GRAPH------------
        //GraphInfo info = ProcessData.readIDNeighborhoodFile(new File(filename1));
        //info.ipToID.clear();
        //origGraph = new Graph(info.idToNeighbors);
		origGraph = new Graph(ProcessData.readTabData(filename1,"\t", selfloopsValue));
        //System.out.println(origGraph.compactStr());

        // read the summary graph
        summaryGraph = readSummaryGraph(filename2);

        System.out.println("Time: " +
                           (System.currentTimeMillis()-start)/1000.0);

        System.out.println("COST " + eps + " "
                           + summaryCost + " "
                           + origGraph.computeCost());
    }

    public Graph readSummaryGraph(String filename)
    {
	System.out.println("Reading Summmary from file: " + filename);
        summaryCost = 0;

        Graph g = new Graph();
        HashMap<Integer, SortedSet> nodeMap =
            new HashMap<Integer, SortedSet> ();
        HashMap<Integer, SortedSet> edges =
            new HashMap<Integer, SortedSet> ();

	try {
	    BufferedReader summary =
                new BufferedReader(new FileReader(filename + ".sum"));
	    BufferedReader corrections;
            if(eps == 0.0)
                corrections = new BufferedReader(new FileReader(filename + ".cor"));
            else
                corrections = new BufferedReader(new FileReader(filename + ".cor." + ((int) (100*eps))));
	    BufferedReader mapping =
                new BufferedReader(new FileReader(filename + ".map"));

            if(eps != 0.0)
                System.out.println(filename + ".cor." + ((int) (100*eps)));


            // some temp variables
	    String line = null;
	    String [] words = null;
            TreeSet<Integer> nodes = null;
            Node a;
            int id, degree;

            // first read supernode to node mapping
	    while (( line = mapping.readLine()) != null){
                words = line.split(" ");
                id = Integer.valueOf(words[0]);
//                 System.out.print(id);
                nodes = new TreeSet<Integer>();
                for (int i = 1; i < words.length; i++){
                    int _id = Integer.valueOf(words[i]);
                    nodes.add(_id);
//                     System.out.print(" " + _id);
                    // add the node to the original graph
                    a = new Node(null, _id, 1);
                    g.addNode(a);
        	}
                nodeMap.put(id, nodes);
            }

            // now read superedges
	    while (( line = summary.readLine()) != null){
                words = line.split(" ");
                id = Integer.valueOf(words[0]);
                degree = Integer.valueOf(words[2]);
                nodes = new TreeSet<Integer>();
                for (int i = 3; i < words.length; i+=2){
                    int _id = Integer.valueOf(words[i]);
                    // now add edges to the graph
                    if(id <= _id){
                        addEdges(id, _id, nodeMap, edges);
                        summaryCost++;
                    }
        	}
            }

            // now read corrections
	    while (( line = corrections.readLine()) != null){
                words = line.split(" ");
                id = Integer.valueOf(words[1]);
                int _id = Integer.valueOf(words[2]);
                // update the correction
                summaryCost++;
                if(words[0].compareTo("+")==0){
                    addEdge(id, _id, edges);
                } else {
                    removeEdge(id, _id, edges);
                }
            }

	} catch (FileNotFoundException e) {
	    e.printStackTrace();
	} catch (IOException e) {
	    e.printStackTrace();
	}

        // make the graph
        for (Integer i : g.nodeMap.keySet()){
            Node a = g.nodeMap.get(i);
            ArrayList<Edge> aedges = new ArrayList<Edge> ();
            SortedSet<Integer> anbr = edges.get(i);
            if(anbr == null){
                System.out.println(i + " " + a);
                //continue;
            }
            for (Integer nid : anbr){
                Node b = g.nodeMap.get(nid);
                aedges.add(new Edge(a, b, 1));
            }
            a.edges = aedges;
        }

//         System.out.println(g.inputFormat());
        return g;
    }

    public void addEdges(int sid1, int sid2,
                         HashMap<Integer, SortedSet> nodeMap,
                         HashMap<Integer, SortedSet> edges)
    {
        SortedSet<Integer> nodes1 = nodeMap.get(sid1);
        SortedSet<Integer> nodes2 = nodeMap.get(sid2);

        if(nodes1 == null){
            System.out.println("Error1 ");
        }
        if(nodes2 == null)
            System.out.println("Error2 ");

        // make edges for all nodes in supernode sid1
        for(Integer i: nodes1){
            SortedSet<Integer> _temp = edges.get(i);
            if(_temp == null){
                _temp = new TreeSet<Integer>();
                edges.put(i, _temp);
            }
            _temp.addAll(nodes2);
        }

        // make edges for all nodes in supernode sid2
        if(sid1 != sid2){
            for(Integer i: nodes2){
                SortedSet<Integer> _temp = edges.get(i);
                if(_temp == null){
                    _temp = new TreeSet<Integer>();
                    edges.put(i, _temp);
                }
                _temp.addAll(nodes1);
            }
        }
    }

    public void addEdge(int id1, int id2,
                         HashMap<Integer, SortedSet> edges)
    {
        SortedSet<Integer> _temp = edges.get(id1);
        if(_temp == null){
            _temp = new TreeSet<Integer>();
            edges.put(id1, _temp);
        }
        _temp.add(id2);

        _temp = edges.get(id2);
        if(_temp == null){
            _temp = new TreeSet<Integer>();
            edges.put(id2, _temp);
        }
        _temp.add(id1);
    }

    public void removeEdge(int id1, int id2,
                         HashMap<Integer, SortedSet> edges)
    {
        SortedSet<Integer> _temp = edges.get(id1);
        _temp.remove(id2);

        _temp = edges.get(id2);
        _temp.remove(id1);
    }

    public boolean compare()
    {
        return compare(origGraph, summaryGraph);
    }

    public boolean compare(Graph g1, Graph g2)
    {
        boolean differ = false;
        if(g1.numNodes != g2.numNodes){
            System.out.println("Size of graphs differ! "
                               + g1.numNodes + " " +
                               + g2.numNodes);
            differ = true;
        }

        for (Integer i : g1.nodeMap.keySet()){
            Node a1 = g1.nodeMap.get(i);
            Node a2 = g2.nodeMap.get(i);
            if(a1 == null){
                System.out.println("Node not present in g1: " + a2);
                differ = true;
            }
            if(a2 == null){
                System.out.println("Node not present in g2" + a1);
                differ = true;
            }
            ArrayList<Integer> aedges = a1.getNeighborIDs();
            ArrayList<Integer> bedges = a2.getNeighborIDs();
            ArrayList<Integer> bedges2 = a2.getNeighborIDs();
            int maxDiff = (int) (eps * aedges.size());
            bedges2.removeAll(aedges);
            aedges.removeAll(bedges);
            if(aedges.size() > maxDiff || bedges2.size() > maxDiff){
                System.out.println("Edges differ for node id " + i);
                System.out.println(maxDiff + " " +
                                   aedges.size() + " " +
                                   bedges2.size());
                System.out.println(a1);
                System.out.println(a2);
                differ = true;
            }
        }

        return differ;
    }
}
