import graph.Edge;
import graph.Graph;
import graph.GraphInfo;
import graph.Node;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

public class ProcessData
{

	public static HashMap<Integer,SortedSet> readTabData(String filename, String sep, Boolean selfloops)
	{
		HashMap<Integer, SortedSet> originalEdges = new HashMap<Integer,SortedSet>();
		HashMap<String,Integer> mapping = new HashMap<String,Integer>();
		int lineNum = 0, id1 = 0, id2 = 0, id=0;
		String id1Str, id2Str;
		String[] edge;
		BufferedReader input = null;
		FileWriter out = null;

		try
		{
			//System.out.println("filename "+filename);
			input = new BufferedReader( new FileReader( new File(filename) ));
			out = new FileWriter(filename+".ids");

			String line = null;
			while (( line = input.readLine()) != null)
			{
				lineNum++;
				//if (lineNum % 100000 == 0)
			//		System.out.println("  Progress: " + lineNum);

				edge = line.split(sep);

				id1Str = edge[0];
				id2Str = edge[1];

				if (mapping.containsKey(id1Str))
					id1 = mapping.get(id1Str);
				else
				{
					out.write(id1Str + sep + id + "\n");
					mapping.put(id1Str, id);
					id1 = id;
					id++;
				}

				if (mapping.containsKey(id2Str))
					id2 = mapping.get(id2Str);
				else
				{
					out.write(id2Str + sep + id + "\n");
					mapping.put(id2Str, id);
					id2 = id;
					id++;
				}

				if (!originalEdges.containsKey(id1))
					originalEdges.put(id1, new TreeSet<Integer>());

				if (!originalEdges.containsKey(id2))
					originalEdges.put(id2, new TreeSet<Integer>());

				originalEdges.get(id1).add(id2);
				originalEdges.get(id2).add(id1);

			}

			out.close();

		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		// Add self-edges.
		if (selfloops != null && selfloops)
		{
			for (String u : mapping.keySet())
			{
	   		 	int uid = mapping.get(u);
			    originalEdges.get(uid).add(uid);
			}
		}

		return originalEdges;
	}

    public static GraphInfo makeIDNeighborhoodFile(Graph g)
    {
    	System.out.println("Making ID Neighborhood File...");
    	HashMap<String,Integer> ipToID = new HashMap<String,Integer>();
    	HashMap<Integer, SortedSet> idToNeighbors = new HashMap<Integer,SortedSet>();

    	for (Integer i : g.nodeMap.keySet())
    	{
    		Node n = g.nodeMap.get(i);
    		idToNeighbors.put(n.id, new TreeSet());

    		ArrayList<Edge> edges = n.edges;

    		for (Edge e : edges)
    			idToNeighbors.get(n.id).add(e.end.id);

    	}

    	return new GraphInfo(ipToID, idToNeighbors);
    }

    public static GraphInfo readIDNeighborhoodFile(File filename)
    {
	System.out.println("Reading ID Neighborhood File...");
	String[] node = null;
	HashMap<String,Integer> ipToID = new HashMap<String,Integer>();
	HashMap<Integer, SortedSet> idToNeighbors = new HashMap<Integer,SortedSet>();
	BufferedReader input = null;
	int id;

	try {
	    input = new BufferedReader( new FileReader(filename) );
	    String line = null;
	    while (( line = input.readLine()) != null)
        {

		    node = line.split(" ");
		    id = Integer.valueOf(node[1]);
		    ipToID.put(node[0], id);
		    idToNeighbors.put(id, new TreeSet());

		    for (int i = 3; i < node.length; i+=2) // Change this to i=i+2 for weights. i++ without.
        	{
			    idToNeighbors.get(id).add(Integer.valueOf(node[i]));
        	}
       }

	} catch (FileNotFoundException e) {
	    e.printStackTrace();
	} catch (IOException e) {
	    e.printStackTrace();
	}

	return new GraphInfo(ipToID, idToNeighbors);
    }


	public static Graph readSummaryData(String filename)
    {
    	Graph g = new Graph();
		System.out.println("Reading Summary...");
		int lineNum = 0, id1, id2;
		int id = 1;
		String[] edge;
		String check;
		BufferedReader input = null;
		HashMap<Integer,Integer> n2n = new HashMap<Integer,Integer>();

	    try
	    {
		    input = new BufferedReader( new FileReader( new File(filename) ));
		    String line = null;
		    while (( line = input.readLine()) != null)
        	{
			    lineNum++;
			    if (lineNum % 100000 == 0)
			    	System.out.println("  Progress: " + lineNum);

			    edge = line.split(" ");

			    id1 = Integer.parseInt(edge[0]);

			    if (n2n.containsKey(id1))
			    {
			    	id1 = n2n.get(id1);
			    }
			    else
			    {
			    	n2n.put(id1, id);
			    	id1 = id;
			    	id++;
			    }


			    if (!g.nodeMap.containsKey(id1))
			    {
			    	g.addNode(id1);
			    }

			    for (int i = 3; i < edge.length; i+=2)
			    {
			    	id2 = Integer.parseInt(edge[i]);

			    	if (n2n.containsKey(id2))
				    {
				    	id2 = n2n.get(id2);
				    }
				    else
				    {
				    	n2n.put(id2, id);
				    	id2 = id;
				    	id++;
				    }

			    	if (id1 > id2)
			    		continue; // Don't want to double count edges.

			    	if (!g.nodeMap.containsKey(id2))
				    {
				    	g.addNode(id2);
				    }

				    g.addEdge(id1, id2, 1);

			    }
        	}

		    System.out.println(n2n.get(-16249 ));

		} catch (FileNotFoundException e) {
		    e.printStackTrace();
		} catch (IOException e) {
		    e.printStackTrace();
		}

		return g;

    }

    public static Graph readCorrectionsData(String filename)
    {
    	Graph g = new Graph();
		System.out.println("Reading Corrections...");
		int lineNum = 0, id1, id2;
		String[] edge;
		String check;
		BufferedReader input = null;

	    try
	    {
		    input = new BufferedReader( new FileReader( new File(filename) ));
		    String line = null;
		    while (( line = input.readLine()) != null)
        	{
			    lineNum++;
			    if (lineNum % 100000 == 0)
			    	System.out.println("  Progress: " + lineNum);

			    edge = line.split(" ");

			    check = edge[0];

			    if (check.equals("+"))
			    	continue; // Only look at positive edges.

			    id1 = Integer.parseInt(edge[1]);
			    id2 = Integer.parseInt(edge[2]);

			    if (!g.nodeMap.containsKey(id1))
			    {
			    	g.addNode(id1);
			    }

			    if (!g.nodeMap.containsKey(id2))
			    {
			    	g.addNode(id2);
			    }

			    g.addEdge(id1, id2, 1);

        	}

		} catch (FileNotFoundException e) {
		    e.printStackTrace();
		} catch (IOException e) {
		    e.printStackTrace();
		}

		return g;

    }

}
