import graph.Edge;
import graph.Graph;
import graph.GraphInfo;
import graph.Node;

import java.io.File;
import java.util.HashSet;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Random;

import options.CmdLineParser;

public class Randomized
{
    static HashSet<Integer> dirtyNodes = new HashSet<Integer>();
    static Random randGenerator = new Random();
    static Graph g;
    static long startTime;
    static int[] U;
    static int[] rIndexU;
    static int uSize;
    static int numNodes;

    public static void main(String[] args)
    {
        startTime = System.currentTimeMillis();

		// get options.
		CmdLineParser parser = new CmdLineParser();
		CmdLineParser.Option selfloops = parser.addBooleanOption('s',"self");

		try { parser.parse(args); }
		catch ( CmdLineParser.OptionException e ) {
            System.err.println(e.getMessage());
            System.err.println("Usage: Randomized [-s,--self] graph_file");
            System.exit(2);
        }

		Boolean selfloopsValue = (Boolean) parser.getOptionValue(selfloops);

        // these are 10 randomly generated numbers,
        // picked as seeds for 10 runs of randomized
        ArrayList<Integer> seeds = new ArrayList<Integer>();
        seeds.add(12345345);
        //seeds.add(578923578);
        //seeds.add(63217890);
        //seeds.add(7238451);
        //seeds.add(48902366);
        //seeds.add(423234);
        //seeds.add(5689043);
        //seeds.add(247391);
        //seeds.add(1431240);

		//-------GETS ARGS------------
		String[] otherArgs = parser.getRemainingArgs();
        String filename = otherArgs[0];
        System.out.println("#Reading data from input file: " + filename);
		if (selfloopsValue != null && selfloopsValue)
			System.out.println("# Adding a self loop to each node...");
        //----------------------------

        //----------READ INPUT FILE, MAKE GRAPH------------
        //GraphInfo info = ProcessData.readIDNeighborhoodFile(new File(filename));
        //info.ipToID.clear();

        // make 10 runs of the algo
        int nseed = 0;
        int snodes = 0, summaryCost = 0, correctionCost = 0;
        for (Integer seed : seeds){
            randGenerator.setSeed(seed);
            //System.out.println("Running with seed: " + nseed + " " + seed);
            // this will keep the originaledges in the graph,
            // needed to print the summary
			g = new Graph(ProcessData.readTabData(filename,"\t",selfloopsValue));
            //g = new Graph(info.idToNeighbors);
            g.eps = -1;
            if(nseed == 0)
                System.out.println(g.compactStr());
	    handleSame();
		//System.out.println(g.compactStr());
	    runRandomized();
	    //System.out.println(g.compactStr());
            g.computeCost();
            snodes += g.nodeMap.size();
            summaryCost += g.summaryCost;
            correctionCost += g.correctionCost;
            nseed++;
            //System.out.println(" Running time: " + runtime());
        }

        snodes/=nseed;
        summaryCost /= nseed;
        correctionCost /= nseed;

        //System.out.println(" # SNODES: " + snodes);
        //System.out.println(" # COST: " + (summaryCost + correctionCost));
        //System.out.println(" # SEDGES: " + summaryCost);
        //System.out.println(" # CORRECTIONS: " + correctionCost);
        //System.out.println(" # TOTAL TIME: " + runtime()/seeds.size());

        g.writeSummary(filename);
		g.printStatistics();
	    System.out.println("\n#Total time (secs): " + runtime());
    }

    public static void runRandomized()
    {
        System.out.println("#Running Randomized...");
        //----------INITIALIZE VARIABLES-----------------------
        HashSet<Node> endIDs = new HashSet<Node>();
        HashSet<Node> cluster = new HashSet<Node>();

        // the U set
        // U[i] = id of some node
        // rIndexU[n+id] = index of node with id (+ve or -ve) in U
        // rIndexU[n+1 - 2n]: original nodes
        // rIndexU[1 - n-1]: snodes (in reverse order)
        // idea is: while picking a node, just remove its entry from U
        // (swap with last element), and make rIndexU[U[i]] = -1
        // for jid = best pair of i, use rIndexU[U[i]]
        // to remove its entry from U

        if(U == null){
            int maxId = -1;
            for(Integer id : g.nodeMap.keySet()){
                maxId = maxId > id ? maxId : id;
            }
            U = new int[maxId+1];
            rIndexU = new int[2*maxId+2];
            numNodes = maxId;
        }

	int _i=1;
	for(Integer i : g.nodeMap.keySet()){
            U[_i] = i;
            rIndexU[numNodes + i] = _i;
	    _i++;
	}
        uSize = g.nodeMap.size();

        // Initialize U.
//         HashSet<Integer> U = new HashSet<Integer>();
//          for (Integer i : g.nodeMap.keySet()){
//             U.add(i);
//         }

        int iterations = 0;
        while (uSize > 0){
            iterations++;

            int i = pickRandomNode();
            Node n = g.nodeMap.get(i);
            if(n==null){
                // some ids may not be present :(
                //sanityCheck();
                continue;
            }

            // Get 2-hop neighbors of n.
            endIDs.clear();
            cluster.clear();
            for (Edge neighbors : n.edges)
                endIDs.addAll(neighbors.end.getNeighbors());

            // Find 2-hop neighbor which gives greatest cost reduction.
            //double _maxCost = -10; // JUST FOR ADAM!!
            double _maxCost = 0;
            Node _maxNode = null;
            for (Node neighbor : endIDs){
                if (neighbor.id == n.id)
                    continue;

                if (rIndexU[numNodes + neighbor.id] == -1)
                    continue;

                double temp = getCost(n,neighbor);
                //System.out.println("Cost: " + n.id + " " + neighbor.id + " " + temp);
                if (temp > _maxCost){
                    _maxCost = temp;
                    _maxNode = neighbor;
                }
            }

            // Check if positive cost reduction found.
            //if (_maxCost > -10){ // JUST FOR ADAM!!!
            if (_maxCost > 0.01){
                cluster.add(n);
                cluster.add(_maxNode);
                //System.out.println("Clustering: " + n.id + " " + _maxNode.id + " " + uSize);
                int clusterID = g.makeClusterFromNodes2(cluster);
                replaceInU(_maxNode.id, clusterID);
//             } else {
//                 System.out.println("Finished: " + n.id + " " + uSize);
            }

//             if (iterations % 1000 == 0){
//                 System.out.println("Running time " + uSize
//                                    + " " + runtime());
//                 //sanityCheck();
//             }

//             printU();
//             printrIndexU();
//            sanityCheck();
         }

//         System.out.println(g.compactStr());
    }

    public static int pickRandomNode()
    {
        // return a random node, and ALSO remove is from U and rIndexU
        int randID;
        if(uSize == 1)
            randID = 1;
        else
            randID = randGenerator.nextInt(uSize-1) + 1;
        int id = U[randID];
        rIndexU[numNodes + id] = -1;
        U[randID] = U[uSize];
        rIndexU[numNodes + U[randID]] = randID;
        uSize--;

        //System.out.println("Picked " + id + " " + uSize);

        return id;
    }

//     public static void removeFromU(int i)
//     {
//         System.out.println("Removing: " + i);
//         int id = rIndexU[numNodes + i];
//         rIndexU[numNodes + i] = -1;
//         System.out.println("Removing2: " + id);
//         U[id] = U[uSize];
//         rIndexU[numNodes + U[id]] = id;
//         uSize--;
//     }

    public static void replaceInU(int i, int j)
    {
        //System.out.println("Replacing: " + i + " " + j);
        // put j in i's entry in U
        int id = rIndexU[numNodes + i];
        rIndexU[numNodes + j] = id;
        rIndexU[numNodes + i] = -1;
//         if(id < 1 || id > uSize){
//             System.out.println("Replacing: " + i + " " + j + " " + id);
//             sanityCheck();
//         }
        U[id] = j;
    }

    public static double getCost(Node a, Node b)
    {
	// previous costs might have changed;
	// just recompute everything
	a.cost = -1;
	b.cost = -1;
        double costPrev = g.getCost(a) + g.getCost(b);
        int j = b.id;
        if(a.existsEdge(j)){
            // edge between i and j
            // count cost of edge (i,j) only once
            costPrev -= a.getCost(j);
        }
        double costNew = g.getCost(a, b);
        return (costPrev - costNew)/((double)costPrev);
    }

    public static void sanityCheck()
    {
        for(int i=1; i<=uSize; i++){
            if(rIndexU[numNodes + U[i]] != i)
                System.out.println("ERROR " + i + " " + U[i]);
        }
    }

//     public static void printU()
//     {
//         for(int i=1; i<=uSize; i++){
//             System.out.print(U[i] + " ");
//         }
//         System.out.println();
//     }

//     public static void printrIndexU()
//     {
//         for(int i=1; i<=2*numNodes; i++){
//             if(rIndexU[i] != -1)
//                 System.out.println((i-numNodes) + " " + rIndexU[i]);
//         }
//     }


    public static void handleSame()
    {
	// Visit every node with this x value.
        System.out.println("\n#Running handle same...");
        int n = g.numNodes;
	for (int i=0; i<n; i++){
	    Node a = g.nodeMap.get(i);
	    // It has already been clustered.
	    if (a == null)
		continue;

	    HashSet<Node> cluster = new HashSet<Node>();
	    HashMap<Integer,Double> profileSim =
		Utilities.make2HopSimilarity(a);

	    cluster.add(a);

	    for (Integer i1 : profileSim.keySet()){
		if (profileSim.get(i1) > .9999){
		    cluster.add(g.nodeMap.get(i1));
		}
	    }
	    if (cluster.size() > 1)
		g.makeClusterFromNodes2(cluster);
	}

        //System.out.println("handleSame done!");
	return;
    }


    private static double runtime()
    {
        return (System.currentTimeMillis()-startTime)/1000.0;
    }
}
