import graph.Edge;
import graph.Graph;
import graph.Node;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.SortedSet;


public class Utilities 
{
    /**
     * Takes as input the id to neighbors graph file and returns the corresponding Graph.
     * @param idToNeighbors
     * @return
     */
    public static Graph MakeGraphFromIDFile(HashMap<Integer, SortedSet> idToNeighbors)
    {
	System.out.println("Making Graph From ID File...");
	Graph g = new Graph();
	Iterator it;
		
	// TODO: Optimize: add them all once, then add edges without checking if nodes exist.
	for (Integer i : idToNeighbors.keySet()){
	    g.addNode(i);		
	}
	
	for (Integer i : idToNeighbors.keySet())
	{
	    for (it = idToNeighbors.get(i).iterator(); it.hasNext();)
	    {
			Integer j = (Integer)it.next();
			if(i <= j)
			    g.addEdge(i, j, 1);
		}
	}
				
	return g;
    }

    /**
     * For each node, computes its maximum similarity with any other node in the graph. Puts sims into a HashMap sims
     * which maps sim to count (where count is the number of pairs of nodes which have maximum similarity sim with any other node).
     * A distribution. 
     * @param g
     */
    public static void find2HopSimDist(Graph g)
    {
	HashMap<Double,Integer> sims = new HashMap<Double,Integer>();
	HashMap<Integer,Double> profileSim = new HashMap<Integer,Double>();
	Node profile;
	int count = 0;
	double maxSim;
		
	for (Integer i : g.nodeMap.keySet())
	    {
		count++;
		if (count % 1000 == 0)
		    System.out.println(" Progress: " + count);
			
		profile = g.nodeMap.get(i);
		profileSim = make2HopSimilarity(profile);
			
		//find maxSim
		maxSim = 0.0;
			
		for (Integer j : profileSim.keySet())
		    {
			double sim = profileSim.get(j);
			if (sim > maxSim && j.compareTo(profile.id) != 0)
			    {
				maxSim = sim;
			    }				
		    }
			
		if (!sims.containsKey(maxSim))
		    sims.put(maxSim, 1);
		else
		    sims.put(maxSim,sims.get(maxSim)+1);
	    }
		
	ArrayList<Double> keys = new ArrayList<Double>(sims.keySet());
	Collections.sort(keys);
		
	for (Double sim : keys)
	    {
		System.out.println(sim + " " + sims.get((sim)));
	    }
    }
	
    /**
     * Computes the similarity between node and all 2-hop distance neighbors. All other nodes
     * have similarity 0 with node. 
     * @param node
     * @return
     */
    public static HashMap<Integer,Double> make2HopSimilarity(Node node)
    {
	HashMap<Integer,Double> simMap = new HashMap<Integer,Double>();
	HashMap<Integer,Double> numMap = new HashMap<Integer,Double>();
	HashMap<Integer,Double> denMap = new HashMap<Integer,Double>();
	ArrayList<Edge> neighbors = node.edges;
	ArrayList<Edge> neighborsNeighbors;
	double sim;
	Edge e,ee;
		
	double wtNode1 = node.nodeWeight;
	double sumFrac = node.sumEdgeWeight/wtNode1;
	for (int i = 0; i < neighbors.size(); i++)
	    {
		e = neighbors.get(i);
		double fraction1 = e.weight/wtNode1;
		neighborsNeighbors = e.end.edges;

		// numMap = \sum{ min(a_i, b_i) }
		// a_i = w(u,x)/w(u) , etc...
		// denMap = \sum{ a_i } + \sum{ b_i }
		// simMap = numMap.get(i) / (denMap.get(i) - numMap.get(i))
		// USE: max(a,b) = a + b - min(a,b)
		//
		for (int j = 0; j < neighborsNeighbors.size(); j++)
		    {
			ee = neighborsNeighbors.get(j);
			double wtNode2 = ee.end.nodeWeight;
			int eeid = ee.end.id;
			double fraction2 = ee.weight/wtNode2;
			double min = Math.min(fraction1,fraction2);
			if (numMap.containsKey(eeid))
			    {
				min += numMap.get(eeid);
				numMap.remove(eeid);
				numMap.put(eeid, min);
			    }else{
				numMap.put(eeid, min);
				double sum = sumFrac + 
				    ee.end.sumEdgeWeight/wtNode2;
				denMap.put(eeid, sum);
			    }
		    }
	    }
		
	for(Integer i : numMap.keySet()){
	    simMap.put(i, numMap.get(i)/(denMap.get(i) - numMap.get(i)));
	}

	// 		HashMap<Integer,Double> simMap2 = make2HopSimilarity_old(node);
	// 		for(Integer i : simMap.keySet()){
	// 		    if(Math.abs((simMap.get(i) - simMap2.get(i))) > .05){
	// 			System.out.println("Not Same: " 
	// 					   + node.getID() + " " + i + " " 
	// 					   + simMap.get(i) + " " + simMap2.get(i));
	// 			System.out.println(computeSimilarity3(node, g.nodeMap.get(i)));
	// 			System.out.println(denMap.get(i) - numMap.get(i));
	// 			System.console().readLine();
	// 		    }
	// 		}
	return simMap;
		
    }
	
    public static ArrayList<Node> findArrayListIntersection(ArrayList<Node> n1, ArrayList<Node> n2)
    {
	ArrayList<Node> intersection = new ArrayList<Node>();
	int n1Index = 0;
	int n2Index = 0;
      
	while ((n1Index < n1.size()) && (n2Index < n2.size())) 
	    {
		if(n1.get(n1Index).id < n2.get(n2Index).id) 
 		    {
			n1Index++;
 		    }
		else if(n1.get(n1Index).id > n2.get(n2Index).id)
		    {
			n2Index++;
		    }
		else 
 		    {
 		    	intersection.add(n1.get(n1Index));
 		    	n1Index++;
 		    	n2Index++;
 		    }
	    }
		
	return intersection;
    }	
	
    public static ArrayList<Node> mergeArrayLists(ArrayList<Node> n1, ArrayList<Node> n2)
    {
	ArrayList<Node> result = new ArrayList<Node>(n1.size() + n2.size());
	int n1Index = 0;
	int n2Index = 0;
      
	while ((n1Index < n1.size()) && (n2Index < n2.size())) 
	    {
		if(n1.get(n1Index).id < n2.get(n2Index).id) 
		    {
			result.add(n1.get(n1Index++));
		    }
		else 
		    {
		    	result.add(n2.get(n2Index++));
		    }
	    }
		
	// Copy any remaining parts of each vector. 
	while(n1Index < n1.size()) 
	    {
		result.add(n1.get(n1Index++));
	    }
		  
	while(n2Index < n2.size()) 
	    {
		result.add(n2.get(n2Index++));
	    }
		
	return result;
    }
    
    public static ArrayList<Node> mergeAndRemoveDups(ArrayList<Node> n1, ArrayList<Node> n2)
    {
	ArrayList<Node> result = new ArrayList<Node>(n1.size() + n2.size());
	int n1Index = 0;
	int n2Index = 0;
      
	while ((n1Index < n1.size()) && (n2Index < n2.size())) 
	    {
		if(n1.get(n1Index).id < n2.get(n2Index).id) 
		    {
			result.add(n1.get(n1Index++));
		    }
		else if(n1.get(n1Index).id == n2.get(n2Index).id)
		{
			result.add(n1.get(n1Index));
			n1Index++;
			n2Index++;
		}
		else 
		    {
		    	result.add(n2.get(n2Index++));
		    }
	    }
		
	// Copy any remaining parts of each vector. 
	while(n1Index < n1.size()) 
	    {
		result.add(n1.get(n1Index++));
	    }
		  
	while(n2Index < n2.size()) 
	    {
		result.add(n2.get(n2Index++));
	    }
		
	return result;
    }
	
	



}
