package graph;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.SortedSet;
import java.util.Collections;

public class Graph
{
    public TreeMap<Integer,Node> nodeMap;
    public int numNodes;
    public int numEdges;
    public Integer numClusters;
    public Integer numSingleNodes;
    public HashMap<Integer, SortedSet> originalEdges;
    public double eps;
    public int summaryCost;
    public int correctionCost;

    public Graph()
    {
	nodeMap = new TreeMap<Integer,Node>();
	numNodes = 0;
	numEdges = 0;
	numClusters = 0;
	numSingleNodes = 0;
	originalEdges = null;
    }

    public Graph(HashMap<Integer, SortedSet> idToNeighbors)
    {
	//System.out.println("Making Graph From ID File...");
	nodeMap = new TreeMap<Integer,Node>();
	numNodes = 0;
	numEdges = 0;
	numClusters = 0;
	numSingleNodes = 0;
	originalEdges = idToNeighbors;
	Iterator it;

	// TODO: Optimize: add them all once, then add edges without checking if nodes exist.
	for (Integer i : idToNeighbors.keySet()){
	    addNode(i);
	}

	for (Integer i : idToNeighbors.keySet()){
	    Node n1 = nodeMap.get(i);
	    for (it = idToNeighbors.get(i).iterator(); it.hasNext();){
		Integer j = (Integer)it.next();
		if(i <= j){
		    Node n2 = nodeMap.get(j);
                    n1.addEdge(n2, 1);
                    n2.addEdge(n1, 1);
                    numEdges++;
		}
	    }
	}

    }


    // TODO: Now, each single node has an empty TreeSet hanging around. Will take space.
    public Node addNode(HashSet<Node> nodes, Integer id, int nodeWeight)
    {


	// If in the HashSet there is a cluster,
	//then we need to individually add those nodes, not the -x.
	Node n = new Node(new TreeSet<Integer>(),id,nodeWeight);
	nodeMap.put(id, n);
	numNodes++;

	// If nodes is null, then we're just adding
	// a single node. Else, add cluster node.
	for (Iterator i = nodes.iterator(); i.hasNext();){
	    Node a = (Node) i.next();
	    numNodes--;

	    // If a is a cluster node, go through cluster
	    // and add ids. Could also check if cluster by negative id.
	    if (a.nodes == null){
		n.nodes.add(a.id);
		numSingleNodes--;
	    } else {
		n.nodes.addAll(a.nodes);
	    }
	}
        return n;
    }

    // Cluster constructor.
    public void addNode(Node a, Node b, Integer id, int nodeWeight)
    {
		// If in the HashSet there is a cluster,
		// then we need to individually add those nodes, not the -x.
		Node _node = new Node(new TreeSet<Integer>(), id, nodeWeight);
		nodeMap.put(id, _node);
		numNodes++;

		numNodes--;
		if (a.nodes == null)
		{
		    _node.nodes.add(a.id);
		    numSingleNodes--;
		}
		else
		{
		    _node.nodes.addAll(a.nodes);
		}

		numNodes--;
		if (b.nodes == null)
		{
		    _node.nodes.add(b.id);
		    numSingleNodes--;
		}
		else
		{
		    _node.nodes.addAll(b.nodes);
		}
    }

    // Single constructor.
    public void addNode(Integer id)
    {
		if (!nodeMap.containsKey(id))
		{
			nodeMap.put(id, new Node(null,id,1));
			numNodes++;
			numSingleNodes++;
		}
    }

    // add an empty node to g
    // will add edges explicitely
    public void addNode(Node a)
    {
        if (nodeMap.containsKey(a.id)){
            System.err.println("ERROR: node already exists: " + a);
        }
        nodeMap.put(a.id, a);
        numNodes++;
        if(a.nodes == null)
            numSingleNodes++;
    }

    public void addEdge(int id1, int id2, int weight)
    {
        Node n1 = nodeMap.get(id1);
        Node n2 = nodeMap.get(id2);

        if (n2 == null || n1 == null){
            return;
        }

        if (id1 == id2){
            n1.addEdge(n2,weight);
            numEdges++;
            return;
        }

        n1.addEdge(n2,weight);
        n2.addEdge(n1,weight);
        numEdges++;
    }

    public void addEdge(Node n1, Node n2, int weight)
    {
        if (n2 == null || n1 == null){
            return;
        }

        if (n1.id == n2.id){
            n1.addEdge(n2,weight);
            numEdges++;
            return;
        }

        n1.addEdge(n2,weight);
        n2.addEdge(n1,weight);
        numEdges++;
    }

    // Remove n from graph and all its incident edges.
    public void removeNode(Node n)
    {
        // Remove edges from neighbors to n.
        ArrayList<Edge> edges = n.edges;
        Edge e;
        for (Iterator i = edges.iterator(); i.hasNext();){
            e = (Edge) i.next();
            Node neighbor = e.end;

            // Self-edge, can't modify the iterator,
            // so remove from the iterator.
            // so DO NOTHING !1
            if (neighbor.id != n.id)
            {
            	neighbor.removeEdge(new Edge(neighbor,n,1));
            }

            // 	    }

	}

	// update number of edges in summary
	numEdges-=n.edges.size();

	// Remove its own list of edges and itself from the map.
	// Might not have to do this.
	n.edges.clear();
	nodeMap.remove(n.id);

	// Don't need to decrement numNodes here.
	// Taken care of in addNode (when a cluster is formed, we add
	// one node for the cluster, and decrement the
	// count for all nodes being added to the cluster.

    }

    public int makeClusterFromNodes2(Node a, Node b)
    {
	//System.out.println("in makeClusterFromNodes2b");
	// initialize variables, add supernode
	int clusterID = (++numClusters)*-1;
	int edgeWeight;
	Node node, neighbor;

	addNode(a, b, clusterID, (int) (a.nodeWeight + b.nodeWeight));

	//--------------------------------------------------------------

	/* For each node in the cluster, find all its neighbors.
	 *  If neighbor is in cluster add weight onto selfEdgeWeight.
	 *  If not, find common edges between neighbors and cluster nodes. The weights on those
	 *	edges will be the weight of the (neighbor,cluster) links.
	 */
	int ai=0, bi=0;
	ArrayList<Edge> aNbr = a.edges;
	ArrayList<Edge> bNbr = b.edges;
	while(ai < aNbr.size() || bi < bNbr.size())
	{
	    if(ai == aNbr.size()){
		// aNbr is done
		int _id = bNbr.get(bi).end.id;
		int _wt = bNbr.get(bi).weight;
		if(!(_id == a.id || _id == b.id))
		    addEdge(clusterID, _id, _wt);
		bi++;
		continue;
	    }
	    if(bi == bNbr.size()){
		// bNbr is done
		int _id = aNbr.get(ai).end.id;
		int _wt = aNbr.get(ai).weight;
		if(!(_id == a.id || _id == b.id))
		    addEdge(clusterID, _id, _wt);
		ai++;
		continue;
	    }

	    int _id1 = aNbr.get(ai).end.id;
	    int _id2 = bNbr.get(bi).end.id;
	    int _wt = 0;
	    int _id = Math.min(_id1, _id2);
	    if(_id1 < _id2){
		// aNbr is smaller
		_wt = aNbr.get(ai).weight;
		ai++;
	    }else if(_id1 > _id2){
		// bNbr is smaller
		_wt = bNbr.get(bi).weight;
		bi++;
	    } else{
		// they are equal
		_wt = aNbr.get(ai).weight + bNbr.get(bi).weight;
		bi++;
		ai++;
	    }
	    if(!(_id == a.id || _id == b.id))
		addEdge(clusterID, _id, _wt);

	}

	// handle Self-edge
	int selfEdgeWeight = a.getEdgeWeight(a) +
	    a.getEdgeWeight(b) + b.getEdgeWeight(b);
	if (selfEdgeWeight > 0){
	    addEdge(clusterID, clusterID, selfEdgeWeight);
	}

	//--------REMOVE CLUSTERED NODES FROM GRAPH-----------------------
	removeNode(a);
	removeNode(b);
	//----------------------------------------------------------------

	return clusterID;
    }


    /**
     * Returns 1 if the cluster decreases the cost.
     * 		   0 if the cluster increases the cost.
     * @param cluster Nodes to cluster.
     */
    public int makeClusterFromNodesCheck(HashSet<Node> cluster)
    {
		//System.out.println("in makeClusterFromNodesCheck");
    	// Find all neighbors of cluster.
        HashSet<Node> endIDs = new HashSet<Node>();
        HashSet<NodePair> edgesToAdd = new HashSet<NodePair>();
        int clusterWeight = 0;
        int nodesCost = 0;
        for (Node node : cluster)
        {
        	nodesCost += getCost(node);
            clusterWeight += node.nodeWeight;
            endIDs.addAll(node.getNeighbors());
        }

        // Add supernode.
        int clusterID = (++numClusters)*-1;
        Node c = addNode(cluster, clusterID, clusterWeight);

        for (Node neighbor : endIDs){
            if (cluster.contains(neighbor)){
                // self edge
                continue;
            }
            // Find the sum of the weights of the edges from neighbor into all nodes in the cluster.
            // If neighbor is a cluster then don't just add the edge to -x, but add all the nodes in -x into the cluster.
            int edgeWeight = sumOfWeights(neighbor, cluster);
            c.addEdge(neighbor, edgeWeight); // one-way adding of the edge. (addEdge(c, neighbor, edgeWeight);)
            edgesToAdd.add(new NodePair(c.id, neighbor.id, edgeWeight));
        }

        // Self-edges.
        int selfEdgeWeight = getSelfWeight(cluster);
        if (selfEdgeWeight > 0){
            addEdge(c, c, selfEdgeWeight);
        }

        //System.out.println("Cluster cost: " + c.cost);
        //System.out.println("Nodes cost: " + nodesCost);

        // Now check the cost of c.
        if (getCost(c) > nodesCost)
        {
        	//System.out.println("NO");
        	// Revert.
        	numClusters--;
        	nodeMap.get(c.id).edges.clear();
        	nodeMap.remove(c.id);
        	return 0;
        }
        else
        {
        	//System.out.println("YES");
        	// Add in edges to c from neighbors.
        	for (NodePair np : edgesToAdd)
        		nodeMap.get(np.id2).addEdge(c, (int) np.cost);

        	// Remove clustered nodes from graph.
            for (Node n : cluster)
                removeNode(n);

            return 1;
        }
	}

    /**
     * 1. Remove nodes in cluster from graph, along with all incident edges.
     * 2. Create supernode: node weight and edge weights.
     * @param clusterP1
     */
    public int makeClusterFromNodes2(HashSet<Node> cluster)
    {
        // Find all neighbors of cluster
        HashSet<Node> endIDs = new HashSet<Node>();
        int clusterWeight = 0;
        for (Iterator i = cluster.iterator(); i.hasNext();){
            Node node = (Node) i.next();
            clusterWeight += node.nodeWeight;
            endIDs.addAll(node.getNeighbors());
        }

        // add supernode
        int clusterID = (++numClusters)*-1;
        Node c = addNode(cluster, clusterID, clusterWeight);

        for (Node neighbor : endIDs){
            if (cluster.contains(neighbor)){
                // self edge
                continue;
            }
            // Find the sum of the weights
            // of the edges from neighbor
            // into all nodes in the cluster.
            // If neighbor is a cluster then
            // don't just add the edge to -x,
            // but add all the nodes in -x into the cluster.
            int edgeWeight = sumOfWeights(neighbor, cluster);
//System.out.println(c + " " + neighbor + " " + edgeWeight);
            addEdge(c, neighbor, edgeWeight);
//             addEdge(clusterID, neighbor.id, edgeWeight);
        }

        // handle self edges
        // add weights (a,b) for a>=b
        // otherwise we over estimate the weights
        int selfEdgeWeight = getSelfWeight(cluster);
        if (selfEdgeWeight > 0){
//System.out.println(c + " " + selfEdgeWeight);
            addEdge(c, c, selfEdgeWeight);
        }

        // remove clustered nodes from graph
        for (Iterator i = cluster.iterator(); i.hasNext();){
            removeNode((Node) i.next());
        }

	return clusterID;
    }

    public int sumOfWeights2(Node from, Node toA, Node toB)
    {
	//System.out.println("in sumOfWeights2");
	ArrayList<Edge> edges = from.edges;
	int edgeWeight = 0;
	Edge e;
	for (int i = 0; i < edges.size(); i++)
	    {
		e = edges.get(i);
		if (e.end.id == toA.id || e.end.id == toB.id)
		    edgeWeight += e.weight;
		else if (e.end.id > toA.id && e.end.id > toB.id)
		    return edgeWeight;
	    }

	return edgeWeight;
    }

    public int sumOfWeights(Node neighbor, HashSet<Node> cluster)
    {
	ArrayList<Edge> edges = neighbor.edges;
	int edgeWeight = 0;
	Edge e;

	for (int i = 0; i < edges.size(); i++){
	    e = edges.get(i);
	    if (cluster.contains(e.end))
		edgeWeight += e.weight;
	}

	return edgeWeight;
    }

    public int getSelfWeight(HashSet<Node> cluster)
    {
		int edgeWeight = 0;
		for (Node a : cluster)
		{
		    for (Node b : cluster)
		    {
				if(a.id > b.id)
				    continue;
				edgeWeight += a.getEdgeWeight(b);
				//System.out.println(a.id + " " + b.id + " " + a.getEdgeWeight(b));
		    }
		}

		return edgeWeight;
    }



    public int computeEdgeDifference()
    {
	//System.out.println("in computeEdgeDifference");
	ArrayList<Edge> edges = new ArrayList<Edge>();
	int edgeDiff = 0, _totalEdges=0;

	for (Integer i : nodeMap.keySet()){
            Node a = nodeMap.get(i);
	    edges = a.edges;
            int _i = i;

	    for (Edge e : edges){
		// do it for each edge only once!!
		if (e.end.id < _i)
		    continue;

		if (e.end.id == _i &&
		    a.nodeWeight > 1)
		    // ADDED LAST CLAUSE
		    _totalEdges = a.nodeWeight*
			(e.end.nodeWeight+1)/2;
		else
		    _totalEdges = a.nodeWeight*
			e.end.nodeWeight;

		edgeDiff += Math.min(_totalEdges - e.weight, e.weight);
	    }
	}

	return edgeDiff;
    }

    public int computeCost()
    {
	ArrayList<Edge> edges = new ArrayList<Edge>();
	int cost = 0, _totalEdges=0;

        summaryCost = 0;
        correctionCost = 0;
	for (Integer i : nodeMap.keySet()){
	    edges = nodeMap.get(i).edges;

	    for (Edge e : edges){
		// do it for each edge only once!!
		if (e.end.id < i)
		    continue;

		if (e.end.id == i)
		    // ADDED LAST CLAUSE
		    _totalEdges = e.end.nodeWeight*(e.end.nodeWeight+1)/2;
		else
		    _totalEdges = e.start.nodeWeight*e.end.nodeWeight;

                if(2*e.weight > _totalEdges + 1){
                    summaryCost++;
                    correctionCost += _totalEdges - e.weight;
                } else {
                    correctionCost += e.weight;
                }
	    }
	}

        cost = summaryCost + correctionCost;
	return cost;
    }

    // return cost of node index i
    public int getCost(int i)
    {
	//System.out.println("in getCost1");
        Node a = (Node) nodeMap.get(i);
        return getCost(a);
    }

    // return cost of node a
    public int getCost(Node a)
    {
        if(a.cost >= 0)
            return a.cost;

        if(eps >= 0)
            return getCostApprox(a);

//         int apxcost = getCostApprox(a);

        ArrayList<Edge> edges = a.edges;
        int cost = 0, _totalEdges=0;
        for (Edge e : edges){
            if (e.end.id == a.id)
                _totalEdges = (e.end.nodeWeight*(e.end.nodeWeight+1))/2;
            else
                _totalEdges = e.start.nodeWeight*e.end.nodeWeight;
            cost += Math.min(_totalEdges - e.weight + 1, e.weight);
        }

//         if(cost != apxcost){
//             System.out.println("Cost differ " + apxcost + " " + cost);
//             System.out.println(a);
//             for (Edge e : edges){
//                 if (e.end.id == a.id)
//                     _totalEdges = (e.end.nodeWeight*(e.end.nodeWeight+1))/2;
//                 else
//                     _totalEdges = e.start.nodeWeight*e.end.nodeWeight;
//                 System.out.println(e.end.id + " " +
//                                    Math.min(_totalEdges - e.weight + 1, e.weight));
//             }
//             System.exit(0);
//         }

        // keep this node's cost in buffer
        a.cost = cost;
        return cost;
    }

    // return cost of the union of node index i and j
    public int getCost(int index1, int index2)
    {
	//System.out.println("in getCost3");
	Node a = nodeMap.get(index1);
	Node b = nodeMap.get(index2);
        return getCost(a, b);
    }

    // return cost of the union of node index i and j
    // try to optimize
    public int getCost(Node a, Node b)
    {
        if(eps >= 0)
            return getCostApprox(a, b);

//         int apxcost = getCostApprox(a, b);

	int cost = 0, _totalEdges=0;
	int wtc = a.nodeWeight + b.nodeWeight;
        int index1 = a.id;
        int index2 = b.id;

	int ai=0, bi=0;
        int aNbrSize = a.edges.size();
        int bNbrSize = b.edges.size();
	Edge aNbr[] = a.getEdgeArray();
	Edge bNbr[] = b.getEdgeArray();

	while(ai < aNbrSize && bi < bNbrSize){
	    int _id1 = aNbr[ai].end.id;
	    int _id2 = bNbr[bi].end.id;
	    int _wt = 0;
	    int _id = Math.min(_id1, _id2);
	    if(_id1 < _id2){
		// aNbr is smaller
		_wt = aNbr[ai].weight;
		_totalEdges = wtc * aNbr[ai].end.nodeWeight;
		ai++;
	    }else if(_id1 > _id2){
		// bNbr is smaller
		_wt = bNbr[bi].weight;
		_totalEdges = wtc * bNbr[bi].end.nodeWeight;
		bi++;
	    } else{
		// they are equal
		_wt = aNbr[ai].weight + bNbr[bi].weight;
		_totalEdges = wtc * aNbr[ai].end.nodeWeight;
		bi++;
		ai++;
	    }
	    if(!(_id == index1 || _id == index2)){
		cost += Math.min(_totalEdges - _wt + 1, _wt);
	    }

	}

	while(bi < bNbrSize){
            // aNbr is done
            int _id = bNbr[bi].end.id;
            int _wt = bNbr[bi].weight;
            if(!(_id == index1 || _id == index2)){
                _totalEdges = wtc * bNbr[bi].end.nodeWeight;
                cost += Math.min(_totalEdges - _wt + 1, _wt);
            }
            bi++;
	}

	while(ai < aNbrSize){
            // bNbr is done
            int _id = aNbr[ai].end.id;
            int _wt = aNbr[ai].weight;
            if(!(_id == index1 || _id == index2)){
                _totalEdges = wtc * aNbr[ai].end.nodeWeight;
                cost += Math.min(_totalEdges - _wt + 1, _wt);
            }
            ai++;
	}

	// handle Self-edge
	int selfEdgeWeight = a.getEdgeWeight(a) +
	    a.getEdgeWeight(b) + b.getEdgeWeight(b);
	if(selfEdgeWeight > 0){
// 	    _totalEdges = wtc*wtc;
	    _totalEdges = (wtc*(wtc+1))/2;
	    cost += Math.min(_totalEdges - selfEdgeWeight + 1, selfEdgeWeight);
	}

//         if(cost != apxcost){
//             System.out.println("Cost differ " + apxcost + " " + cost);
//             System.out.println(a);
//             System.out.println(b);
//             System.exit(0);
//         }

	return cost;
    }


    // return the size |a \intersection b|
    // assume a and b are sorted
    // assume |b| > |a|
    public int intersectionSize(SortedSet<Integer> a, SortedSet<Integer> b)
    {
	//System.out.println("in intersectionSize");
        int size = 0, asize = a.size(), bsize = b.size();
        if(asize == 0 || bsize == 0)
            return size;
        Iterator aiter = a.iterator();
        Iterator biter = b.iterator();
        int ai = ((Integer)aiter.next()).intValue();
        int bi = ((Integer)biter.next()).intValue();
	while(true){
            //System.out.println(ai + " " + bi);
            if(ai == bi){
                size++;
                if(!aiter.hasNext() || !biter.hasNext())
                    break;
                ai = ((Integer)aiter.next()).intValue();
                bi = ((Integer)biter.next()).intValue();
            }
	    else if(ai < bi){
                if(!aiter.hasNext())
                    break;
                ai = ((Integer)aiter.next()).intValue();
            }
            else{
                if(!biter.hasNext())
                    break;
                bi = ((Integer)biter.next()).intValue();
            }
	}

        //System.exit(0);
	return size;
    }


    // return cost of node a, calculate based on epsilon value
    public int getCostApprox(Node a)
    {
        ArrayList<Edge> edges = a.edges;
        SortedSet<Integer> A_u = a.nodes;

        // for each node x, let N_x be original edges
        // let Q be ALL nodes in the snodes to which 'a' has a sedge
        // let P be ALL nodes in the snodes to which 'a' do not have a sedge
        // let C_x be the total number of corrections for x
        // x has + corrections to P, and - corrections to Q
        // C_x = |N_x \intersection P| + |Q - N_x|
        //     = |P_x| + |Q - {N_x - P_x}|
        // (since only {N_x - P_x} can have a common with Q)
        //     = |P_x| + |Q| - |N_x - P_x|
        // (since only Q is a super set of {N_x - P_x})
        //     = 2*|P_x| + |Q| - |N_x|

        // find the P and Q sets
        //ArrayList<Node> superedges = new ArrayList<Node>();
        ArrayList<Integer> _sedgeSizes = new ArrayList<Integer>();
        SortedSet<Integer> PSet = new TreeSet<Integer>();
        SortedSet<Integer> QSet = new TreeSet<Integer>();
        int _selfEdgeCost = 0;
        for (Edge e : edges){
            // this edge is called e = (a,v)
            int _totalEdges;
            Node v = e.end;
            if (v.id == a.id){
                // handle self edge right here
                // not removing due to eps-constraint currently
                _totalEdges = (v.nodeWeight*(v.nodeWeight+1))/2;
                _selfEdgeCost = Math.min(_totalEdges - e.weight + 1, e.weight);
                continue;
            }
            else
                _totalEdges = a.nodeWeight*v.nodeWeight;
            boolean _isPresent = (2*e.weight > _totalEdges + 1);

            if(_isPresent){
                QSet.addAll(v.nodes);
                //superedges.add(v);
                _sedgeSizes.add(v.nodeWeight);
            } else {
                PSet.addAll(v.nodes);
            }
        }

        // count the cost of each edge
        int cost = 0;
        //Collections.sort(PSet);
        int Q = QSet.size();
        int P = PSet.size();
        //HashMap<Integer, Integer> edgesLeft = new HashMap<Integer,Integer> ();
        int _minLeft = Q + P;
        for( Integer x : A_u){
            // find C_x
            SortedSet _NxSet = (SortedSet) originalEdges.get(x);
            int _Nx = _NxSet.size();
            // dont count self edge of x
            int _tempx = intersectionSize(_NxSet, A_u);
            _Nx-=_tempx;
            int _Px = intersectionSize(_NxSet, PSet);
            int _Cx = 2*_Px + Q - _Nx;
            int _maxRemove = (int) (eps*_Nx);
            if(_maxRemove > _Cx){
                int _xedgesLeft = _maxRemove - _Cx;
                _minLeft = _xedgesLeft < _minLeft ? _xedgesLeft : _minLeft;
            }
            else{
                cost += _Cx - _maxRemove;
                _minLeft = 0;
            }
//             if(a.id == _debugid1 || a.id == _debugid2)
//                 System.out.println("minleft " + x + " " + _Cx + " " + cost);

        }

//         if(a.id == _debugid1 || a.id == _debugid2)
//             System.out.println("cost till now " + cost + " " + _selfEdgeCost);

        // add the cost of self-edge
	cost += _selfEdgeCost;

        // now process superedges
        if(_sedgeSizes.size() > 0){
            Collections.sort(_sedgeSizes);
            for (Integer _t : _sedgeSizes){
                if(_t <= _minLeft)
                    _minLeft = _minLeft - _t;
                else
                    cost++;
            }
        }

//         if(a.id == _debugid1 || a.id == _debugid2)
//             System.out.println("returning cost for " + a.id + " "
//                                + a.edges.size() + " " + cost);
        // keep this node's cost in buffer
        a.cost = cost;
        return cost;
    }

    // return cost of node a, calculate based on epsilon value
    public int getCostApprox(Node a, Node b)
    {
        SortedSet<Node> edges = new TreeSet<Node>();
        edges.addAll(a.getNeighbors());
        edges.addAll(b.getNeighbors());

        SortedSet<Integer> A_u = new TreeSet<Integer>();
        A_u.addAll(a.nodes);
        A_u.addAll(b.nodes);
        int _nodeWeight = a.nodeWeight + b.nodeWeight;

        // find the P and Q sets
        //ArrayList<Node> superedges = new ArrayList<Node>();
        ArrayList<Integer> _sedgeSizes = new ArrayList<Integer>();
        SortedSet<Integer> PSet = new TreeSet<Integer>();
        SortedSet<Integer> QSet = new TreeSet<Integer>();
        int _selfEdgeCost = 0;
        for (Node v : edges){
            // this edge is called e = (a,v)
            int _totalEdges;
            //Node v = e.end;
            if (v.id == a.id || v.id == b.id){
//                 if(v.id == a.id){
//                     // handle slef edge when v.id == b.id
//                     continue;
//                 }
                int _weight = a.getEdgeWeight(b)
                    + a.getEdgeWeight(a)
                    + b.getEdgeWeight(b);
                // handle self edge right here
                // not removing due to eps-constraint currently
                _totalEdges = (_nodeWeight*(_nodeWeight+1))/2;
                _selfEdgeCost = Math.min(_totalEdges - _weight + 1, _weight);
                continue;
            }
            int _weight = a.getEdgeWeight(v) + b.getEdgeWeight(v);
            _totalEdges = _nodeWeight*v.nodeWeight;
            boolean _isPresent = (2*_weight > _totalEdges + 1);

            if(_isPresent){
                QSet.addAll(v.nodes);
                //superedges.add(v);
                _sedgeSizes.add(v.nodeWeight);
            } else {
                PSet.addAll(v.nodes);
            }
        }


        // count the cost of each edge
        int cost = 0;
        //Collections.sort(PSet);
        int Q = QSet.size();
        int P = PSet.size();
        //HashMap<Integer, Integer> edgesLeft = new HashMap<Integer,Integer> ();
        int _minLeft = Q + P;
        for( Integer x : A_u){
            // find C_x
            SortedSet _NxSet = (SortedSet) originalEdges.get(x);
            int _Nx = _NxSet.size();
            // dont count self edge of x
            int _tempx = intersectionSize(_NxSet, A_u);
            _Nx-=_tempx;
            int _Px = intersectionSize(_NxSet, PSet);
            int _Cx = 2*_Px + Q - _Nx;
            int _maxRemove = (int) (eps*_Nx);
            if(_maxRemove > _Cx){
                int _xedgesLeft = _maxRemove - _Cx;
                _minLeft = _xedgesLeft < _minLeft ? _xedgesLeft : _minLeft;
            }
            else{
                cost += _Cx - _maxRemove;
                _minLeft = 0;
            }
//             if(a.id == _debugid1 && b.id == _debugid2)
//                 System.out.println("STEP " + x + " "
//                                    + _maxRemove + " " + _Cx + " " + cost);
        }

        // add the cost of self edge
        cost += _selfEdgeCost;

        // now process superedges
        if(_sedgeSizes.size() > 0){
            Collections.sort(_sedgeSizes);
            for (Integer _t : _sedgeSizes){
                if(_t <= _minLeft)
                    _minLeft = _minLeft - _t;
                else
                    cost++;
//                 if(a.id == _debugid1 && b.id == _debugid2)
//                     System.out.println("STEP " + _minLeft + " "
//                                        + _t + " " + _selfEdgeCost + " " + cost);
            }
        }

//         if(a.id == _debugid1 && b.id == _debugid2)
//             System.out.println("returning2 cost for " + a.id + " " + b.id + " "
//                            + edges.size() + " " + cost);
        return cost;
    }


    public void printGraph(GraphInfo info)
    {
	Edge e;

	//int deg, avgDeg = 0, minDeg = 0, maxDeg = 0, size = 0;
	int manualNumEdges = 0;

	System.out.println("\n\nGRAPH INFO: ");
	System.out.println(" # total nodes: " + numNodes);
        System.out.println("# COST: " + computeCost());
	System.out.println(" # single nodes: " + numSingleNodes );
	System.out.println(" # cluster nodes: " + (numNodes - numSingleNodes));
	System.out.println(" # edges: " + numEdges);
	System.out.println(" edge difference: " + computeEdgeDifference());
	System.out.println(" cost = " + computeCost());

	for (Integer i : nodeMap.keySet())
	{
		// Print out nodes within the cluster.
		if (i < 0)
		{
			//avgDeg = 0;
			//minDeg = 0;
			//maxDeg = 0;
			//size = 0;
			System.out.print("\nNode " + i + "[" + (nodeMap.get(i)).nodeWeight + "]" +" { ");
			for (Iterator j = (nodeMap.get(i)).nodes.iterator(); j.hasNext();)
		    {
				//size++;
				/*deg = info.idToNeighbors.get(n.id).size();
				avgDeg += deg;
				if (deg > maxDeg)
				    maxDeg = deg;
				if (size == 1)
				    minDeg = deg;
				else if (deg < minDeg)
				    minDeg = deg; */

				System.out.print(j.next() + ", ");// (" + deg + ") ");
			 }
			//System.out.print("} : avg (" + avgDeg/size + ") min (" + minDeg + ") max (" + maxDeg + ") : \n");
			System.out.println("}: ");

			for (Iterator it = nodeMap.get(i).edges.iterator(); it.hasNext();)
			{
				e = (Edge) it.next();
				System.out.print(e.end.id + " (" + e.weight + ") ");
				if(e.end.id >= i)
				    manualNumEdges++;
			}
			System.out.println();

		}else{
		    		  System.out.print("\nNode " + i + " (" + nodeMap.get(i).edges.size() + "): ");
		    		  for (Edge e1 : nodeMap.get(i).edges)
		    		  {
		    			  System.out.print(" " + e1.end.id + "(" + e1.weight + ") ");
		    		  }
		    		  System.out.println();
		    manualNumEdges+=nodeMap.get(i).edges.size();
		}

		//System.out.println(" # edges recount: " + manualNumEdges);

	}
	System.out.println();

	//System.out.println("\n Manual num edges " + manualNumEdges);
    }

    public String toString()
    {
	String outStr = new String();

	outStr += "\n\nGRAPH INFO: ";
	outStr += " # total nodes: " + numNodes + "\n";
	outStr += " # single nodes: " + numSingleNodes  + "\n";
	outStr += " # cluster nodes: " + (numNodes - numSingleNodes) + "\n";
	outStr += " # edges: " + numEdges + "\n";
	outStr += " edge difference: " + computeEdgeDifference() + "\n";
	outStr += " cost = " + computeCost() + "\n";

	for (Integer i : nodeMap.keySet())
	{
	    Node a = nodeMap.get(i);
	    outStr +=  "Node " + i + " (" + a.nodeWeight + ")";

	    if(a.nodes != null)
	    {
	    	outStr +=  " { ";
	    	for (Iterator j = a.nodes.iterator(); j.hasNext();)
	    	{
	    		outStr += ((Node)j.next()).id + ", ";
	    	}
	    	outStr +=  "}";
	    }

	    outStr += " [" + nodeMap.get(i).edges.size() + "]: ";

	    for (Edge e : nodeMap.get(i).edges)
	    {
	    	outStr += e.end.id + " (" + e.weight + ") ";
	    }
	    outStr +=  "\n";
	}

	return outStr;
    }

    public String inputFormat()
    {
	String outStr = new String();

	for (Integer i : nodeMap.keySet())
	{
	    Node a = nodeMap.get(i);
	    outStr +=  i + " " + i;

	    outStr += " " + nodeMap.get(i).edges.size();

	    for (Edge e : nodeMap.get(i).edges)
	    {
	    	outStr += " " + e.end.id + " " + e.weight;
	    }
	    outStr +=  "\n";
	}

	return outStr;
    }


    public String compactStr()
    {
		String outStr = new String();

		outStr += "# Graph: #nodes=" + numNodes + ", #edges=" + numEdges + ", cost=" + computeCost();

		return outStr;
    }

    public String compactStrOld()
    {
		String outStr = new String();

		outStr += "\nGRAPH: ";
		outStr += " # total nodes: " + numNodes + "\n";
		outStr += " # cluster nodes: " + (numNodes - numSingleNodes) + "\n";
		outStr += " # edges: " + numEdges + "\n";
		outStr += " # cost = " + computeCost() + "\n";

		return outStr;
    }

    public void printStatistics()
    {
        System.out.println("# COST: " + computeCost());
        System.out.println("# SNODES: " + nodeMap.size());
        System.out.println("# SEDGES: " + summaryCost);
        System.out.println("# CORRECTIONS: " + correctionCost);
    }

    public int addCorrections(FileWriter corrections,
                              Edge e, boolean isPresent)
    {
	Node a = e.start;
	Node b = e.end;
	if(a.id > b.id)
	    return 0;

//         System.out.println(e);
//         System.out.println(isPresent);

	int cost=0;
	SortedSet<Integer> aset = a.nodes;
	SortedSet<Integer> bset = b.nodes;
	if(a.nodes == null){
	    aset = new TreeSet();
	    aset.add(a.id);
	}
	if(b.nodes == null){
	    bset = new TreeSet();
	    bset.add(b.id);
	}

        try {
	    if(isPresent){
		// this edge will be present in the graph summary
		// add negative corrections for {AllPair - e}
		cost++;
		for( Integer x : aset){
		    SortedSet xset = (SortedSet) ((TreeSet)bset).clone();
		    xset.removeAll(originalEdges.get(x));
		    for (Iterator j = xset.iterator(); j.hasNext();){
			int y = (Integer) j.next();
			if(a.id != b.id){
			    corrections.write("- " + x + " " + y + "\n");
			    cost++;
			} else if(x.compareTo(y) <= 0){
			    // if self edge, make sure we only add
			    // one correction per original edge
			    corrections.write("- " + x + " " + y + "\n");
			    cost++;
			}
		    }
		}

	    }else{
		// this edge is not present in the graph summary
		// add positive corrections for {e}
		for( Integer x : aset){
		    SortedSet xset = (SortedSet) ((TreeSet)originalEdges.get(x)).clone();
		    xset.retainAll(bset);
		    for (Iterator j = xset.iterator(); j.hasNext();){
			int y = (Integer) j.next();
			if(a.id != b.id){
			    corrections.write("+ " + x + " " + y + "\n");
			    cost++;
			} else if(x.compareTo(y) <= 0){
			    // if self edge, make sure we only add
			    // one correction per original edge
			    corrections.write("+ " + x + " " + y + "\n");
			    cost++;
			}
		    }
		}
	    }
	} catch (IOException e1) {
	    e1.printStackTrace();
	}
	return cost;
    }

    public void writeSummary(String filename)
    {
	System.out.println("\n#Writing compressed representation in file: " + filename + ".{map,cor,sum,ids}");
	Integer id;
	File outFile = new File(filename);

        try {
	    FileWriter summary = new FileWriter(new File(filename + ".sum"));
	    FileWriter corrections = new FileWriter(new File(filename + ".cor"));
	    FileWriter mapping = new FileWriter(new File(filename + ".map"));

	    for (Integer i : nodeMap.keySet()){
		Node a = nodeMap.get(i);
		int _nedges = 0;
		// count the number of edges in graph summary
		for (Edge e : a.edges){
//System.out.println(i+ " " + e.end.id + " " + e.weight + " " + a.nodeWeight);
		    int _totalEdges;
		    if (e.end.id == i)
			_totalEdges = a.nodeWeight*
			    (a.nodeWeight+1)/2;
		    else
			_totalEdges = a.nodeWeight * e.end.nodeWeight;

		    if(2*e.weight > _totalEdges + 1)
			_nedges++;
		}
		// only write nodes with > 0 degree
		if(_nedges != 0){
		    summary.write(i + " " + a.nodeWeight + " ");
		    summary.write(_nedges + " " );
		}

		// write graph and correction summaries
		for (Edge e : a.edges){
		    int _totalEdges;
		    if (e.end.id == i)
			_totalEdges = a.nodeWeight*
			    (a.nodeWeight+1)/2;
		    else
			_totalEdges = a.nodeWeight * e.end.nodeWeight;

		    int cost = Math.min(_totalEdges - e.weight + 1, e.weight);
		    int cost2;
		    if(2*e.weight > _totalEdges + 1){
			summary.write(e.end.id + " " + e.weight + " ");
			cost2 = addCorrections(corrections, e, true);
		    } else {
			cost2 = addCorrections(corrections, e, false);
		    }
		    if(i <= e.end.id && cost != cost2){
			System.out.println("Problem: " + e
					   + " " + cost
					   + " " + cost2
					   + " " + _totalEdges
					   + " " + e.weight);
			System.out.println(a);
			System.out.println(e.end);
		    }
		}
		if(_nedges != 0)
		    summary.write("\n");

		// write node mapping
                mapping.write(i + " ");
		if(a.nodes.size() == 1){
		    mapping.write(i + "\n");
                } else{
		    for (Iterator j = a.nodes.iterator(); j.hasNext();){
			mapping.write(j.next() + " ");
		    }
		    mapping.write("\n");
		}
	    }

	    summary.close();
	    corrections.close();
	    mapping.close();

	} catch (IOException e1) {
	    e1.printStackTrace();
	}
    }



}
