package graph;

import java.util.ArrayList;
import java.util.Collections;
import java.util.TreeSet;
import java.util.Iterator;

public class Node implements Comparable<Node>
{
    public ArrayList<Edge> edges;
    private Edge[] _edgeArray;
    public TreeSet<Integer> nodes;
    public int id;
    public int nodeWeight;
    public int sumEdgeWeight=0;
    public int cost=-1;
	
    // For single nodes, the TreeSet is null. 
    // For cluster nodes, it contains the other nodes.
    public Node(TreeSet<Integer> nodes, int id, int nodeWeight)	
    {
        this.edges = new ArrayList<Edge>();
        this.nodes = nodes;
        if(nodes == null){
            this.nodes = new TreeSet<Integer>();
            this.nodes.add(id);
        }
        this.id = id;
        this.nodeWeight = nodeWeight;		
        sumEdgeWeight=0;
        cost=-1;
    }
	
    //	 Return true if there exists an edge between this and n.
    public boolean existsEdge(Node n)
    {
	for (int i = 0; i < edges.size(); i++){
	    if (edges.get(i).end.id > n.id)
		return false;
	    else if (edges.get(i).end.id == n.id)
		return true;
	}
	
	return false;
    }
	
    //	 Return true if there exists an edge between this and nodes index id.
    public boolean existsEdge(int id)
    {
	Edge toAdd = new Edge(this,new Node(null, id, 1), 1);
	int index = Collections.binarySearch(edges, toAdd);
	return(index >= 0);
    }
    
    public int getEdgeIndex(int id)
    {
    	Edge toAdd = new Edge(this,new Node(null, id, 1), 1);
    	int index = Collections.binarySearch(edges, toAdd);
    	return index;
    }
	
    // binary search.
    public void addEdge(Node n, int weight)
    {
	Edge toAdd = new Edge(this,n,weight);
	int index = Collections.binarySearch(edges, toAdd);
	if (index < 0){
            sumEdgeWeight += weight;
	    edges.add(-index-1, toAdd);
            _edgeArray = null;
	}
    }
	
    public String toString()
    {
        String temp = new String();
        temp += id;
        if(nodes != null){
            temp += " [" + nodes.size() + ":";
            for (Iterator j = nodes.iterator(); j.hasNext();){
                temp += " " + ((Integer)j.next());
            }
            temp += "]";
        }

        temp += " (" + edges.size() + ":";
        for (Edge e : edges){
            temp += " " + e.end.id + "(" + e.weight + ")";
        }
        temp += ")";
        return temp;
    }

    // binary search.
    public void removeEdge(Edge e)
    {
	int index = Collections.binarySearch(edges, e);
	if (index >= 0){
            sumEdgeWeight -= e.weight;
	    edges.remove(index);
            _edgeArray = null;
	}
    }
	
    public ArrayList<Node> getNeighbors()
    {
        ArrayList<Node> neighbors = new ArrayList<Node>();
        for (Edge e : this.edges){
            neighbors.add(e.end);
        }
        return neighbors;
    }
	
    public Edge[] getEdgeArray()
    {
        if(_edgeArray == null){
            _edgeArray = new Edge[edges.size()];
            _edgeArray = edges.toArray(_edgeArray);
        }
        return _edgeArray;
    }
	
    public ArrayList<Integer> getNeighborIDs()
    {
        ArrayList<Integer> neighbors = new ArrayList<Integer>();
        for (Edge e : this.edges){
            neighbors.add(e.end.id);
        }
        
        return neighbors;
    }

    // return cost of edge (this, id)
    public int getCost(int id)
    {
	Edge toAdd = new Edge(this,new Node(null, id, 1), 1);
	int index = Collections.binarySearch(edges, toAdd);
	if (index < 0){
	    return 0;
	}

        Edge e = edges.get(index);
        int _totalEdges = nodeWeight*e.end.nodeWeight;
        return Math.min(_totalEdges - e.weight + 1, e.weight);
    }
	
    /**
     * Returns the weight of the edge between this and e; 0 if the edge does not exist.
     * @param e
     * @return
     */
    public int getEdgeWeight(Node n)
    {
	Edge toAdd = new Edge(this,n, 1);
	int index = Collections.binarySearch(edges, toAdd);
	if (index < 0){
	    return 0;
	}
        return edges.get(index).weight;
    }

    /**
     * Returns the weight of the edge between this and e; 0 if the edge does not exist.
     * @param e
     * @return
     */
    public int getEdgeWeight(int index)
    {
        // CHECK if this inequality is correct 
        // ">" or ">=" ?
        if(index >= edges.size())
            return 0;
        return edges.get(index).weight;
    }
	
    public int compareTo(Node n) 
    {
        if (n.id > this.id)
            return -1;
        if (n.id < this.id)
            return 1;
        return 0;
    }
	
    public boolean equals(Object o)
    {
    	return (this.id == ((Node) o).id);
    }
	
    public int hashCode()
    {
    	return this.id;
    }
	
}
